#include <slang.h>
#include <csignal>
#include <typeinfo>

#include "window_events.hh"
#include "idlemgr.hh"
#include "term.hh"
#include "kbd.hh"

class mainwindow : public window
{
    winptr background;
	enum { stidle, stresizing, stmoving } evmode;
public:
    mainwindow() : evmode(stidle)
    {
        background = new window;
        background->buf.defblock.ch = '.';
        background->buf.defblock.attr = 0x01;
        addchild(background);
    }
    virtual void resize(unsigned x, unsigned y)
    {
        background->resize(x, y);
        window::resize(x, y);
    }
    virtual void redraw()
    {
        SLsig_block_signals();
        window::redraw();
        SLsmg_normal_video();
        for(unsigned y=0; y<height(); ++y)
        {
            SLsmg_gotorc(y,0);
            for(unsigned x=0; x<width(); ++x)
            {
                SLsmg_set_color(buf.buf[y][x].attr);
                SLsmg_write_char(buf.buf[y][x].ch);
            }
        }
        SLsmg_refresh();
        SLsig_unblock_signals();
    }
public:
    void getevent(unsigned mswait)
    {
        event e = getkeyevent(mswait);
        
        switch(evmode)
        {
        	case stidle:
                if(event_key *p = dynamic_cast<event_key *> (&*e))
                {
                	/* printf("It was a key! ");
                	
                	unsigned a, b=p->ch.size();
                	for(a=0; a<b; ++a)
                		if((unsigned char)p->ch[a] < 32)
                			printf("^%c", p->ch[a]+64);
                		else
                			printf("%c", p->ch[a]);
                	printf("\r\n");
                	*/
                	if(p->key == BQT_KEY_F5 && p->shift)
                	{
                		e = new event_resize(0, 0);
                		evmode = stresizing;
                	}
                	else if(p->key == BQT_KEY_F5)
                	{
                		e = new event_move(0, 0);
                		evmode = stmoving;
                	}
                	else if(p->key == 'x' && p->alt)
                	{
                		throw;
                	}
                }
                break;
            case stresizing:
                if(event_key *p = dynamic_cast<event_key *> (&*e))
                {
                	if(p->key == BQT_KEY_ENTER)
                	{
                		e = new event_resize(0, 0);
                		evmode = stidle;
                	}
                	else if(p->key == BQT_KEY_LT)
                		e = new event_resize(-1, 0);
                	else if(p->key == BQT_KEY_RT)
                		e = new event_resize(1, 0);
                	else if(p->key == BQT_KEY_UP)
                		e = new event_resize(0, -1);
                	else if(p->key == BQT_KEY_DN)
                		e = new event_resize(0, 1);
                }
                /* Gurb says 'TOSI KIVOJA CITYKAVEREITA MINULLA, toinen pkillaa ja toinen castaa cityn tyteen webbeja.' */
                break;
            case stmoving:
                if(event_key *p = dynamic_cast<event_key *> (&*e))
                {
                	if(p->key == BQT_KEY_ENTER)
                	{
                		e = new event_move(0, 0);
                		evmode = stidle;
                	}
                	else if(p->key == BQT_KEY_LT)
                		e = new event_move(-1, 0);
                	else if(p->key == BQT_KEY_RT)
                		e = new event_move(1, 0);
                	else if(p->key == BQT_KEY_UP)
                		e = new event_move(0, -1);
                	else if(p->key == BQT_KEY_DN)
                		e = new event_move(0, 1);
                }
                break;
        }
        
        if(typeid(*e) != typeid(event_none))
        	handleevent(e);
    }
};

window *desktop;

static void screeninit(int=0)
{
    static const char *const colours[16] =
    {
        "black","blue","green","cyan","red","magenta","brown","lightgray",
        "gray","brightblue","brightgreen","brightcyan",
        "brightred","brightmagenta","yellow","white"
    };
    
    for(unsigned b=0; b<16; ++b)
        for(unsigned f=0; f<16; ++f)
            SLtt_set_color(b*16+f, NULL,
                const_cast<char *>(colours[f]),
                const_cast<char *>(colours[b]));
                   
    SLsmg_init_smg();
    SLtt_get_screen_size();
    desktop->resize(SLtt_Screen_Cols, SLtt_Screen_Rows);
    SLang_init_tty(0,0,0);
    SLsmg_cls();
}

void TermInit()
{
    desktop = new mainwindow;

    SLang_init_tty(0,0,0);
    
    SLsig_block_signals();
    SLtt_get_terminfo();
    /* SLkp_init assumes that SLtt_get_terminfo has been called. */
    if(SLkp_init() < 0)
    {
        SLsig_unblock_signals();
        return;
    }
    SLsig_unblock_signals();
    signal(SIGWINCH, screeninit);
}

void TermLoop()
{
    screeninit();
    
	try
	{
	    for(;;)
    	{
	    	desktop->redraw();
    		dynamic_cast<mainwindow *>(desktop)->getevent(idlemgr.tick() ? 0 : 1000);
	    }
	}
	catch(...)
	{
		SLsmg_cls();
	}
}

void TermDone()
{
	delete desktop;
}
